@@ -21,7 +21,7 @@ var resolveModule = function(names, mod, root) {
       parent : mod.parent,
       id : mod.id,
       exports : {}
-    }
+    };
   }
   // we need to traverse the path
   var n = names.shift();
@@ -63,6 +63,17 @@ var Couch = {
   },
   compileFunction : function(source, ddoc) {    
     if (!source) throw(["error","not_found","missing function"]);
+
+    var evaluate_function_source = function(source, evalFunction, sandbox) {
+      sandbox = sandbox || {};
+      if(typeof CoffeeScript === "undefined") {
+        return evalFunction(source, sandbox);
+      } else {
+        var coffee = CoffeeScript.compile(source, {bare: true});
+        return evalFunction(coffee, sandbox);
+      }
+    }
+
     try {
       if (sandbox) {
         if (ddoc) {
@@ -88,12 +99,12 @@ var Couch = {
               ddoc._module_cache[newModule.id] = newModule.exports;
             }
             return ddoc._module_cache[newModule.id];
-          }
+          };
           sandbox.require = require;
         }
-        var functionObject = evalcx(source, sandbox);
+        var functionObject = evaluate_function_source(source, evalcx, sandbox);
       } else {
-        var functionObject = eval(source);
+        var functionObject = evaluate_function_source(source, eval);
       }
     } catch (err) {
       throw(["error", "compilation_error", err.toSource() + " (" + source + ")"]);
@@ -119,7 +130,7 @@ var Couch = {
       }
     }
   }
-}
+};
 
 // prints the object as JSON, and rescues and logs any toJSON() related errors
 function respond(obj) {
