@@ -141,7 +141,7 @@ function setupAdminParty(fun) {
           success : function() {
             removeAdmins(confs, doneFun);
           }
-        }, "admins", remove[0], null);        
+        }, "admins", remove[0], null);
       } else {
         doneFun();
       }
@@ -238,13 +238,13 @@ function saveTestReport(report) {
       report.db = db_info;
       $.couch.info({success : function(node_info) {
         report.node = node_info;
-        db.saveDoc(report);        
+        db.saveDoc(report);
       }});
     };
     var createDb = function() {
       db.create({success: function() {
-        db.info({success:saveReport});        
-      }});    
+        db.info({success:saveReport});
+      }});
     };
     db.info({error: createDb, success:saveReport});
   }
@@ -313,6 +313,11 @@ function T(arg1, arg2, testName) {
   }
 }
 
+function TIsnull(actual, testName) {
+  T(actual === null, "expected 'null', got '"
+    + repr(actual) + "'", testName);
+}
+
 function TEquals(expected, actual, testName) {
   T(equals(expected, actual), "expected '" + repr(expected) +
     "', got '" + repr(actual) + "'", testName);
@@ -414,9 +419,22 @@ function waitForSuccess(fun, tag) {
 
 function waitForRestart() {
   var waiting = true;
-  while (waiting) {
+  // Wait for the server to go down but don't
+  // wait too long because we might miss the
+  // the unavailable period.
+  var count = 25;
+  while (waiting && count > 0) {
+    count--;
     try {
       CouchDB.request("GET", "/");
+    } catch(e) {
+      waiting = false;
+    }
+  }
+  // Wait for it to come back up
+  waiting = true;
+  while (waiting) {
+    try {
       CouchDB.request("GET", "/");
       waiting = false;
     } catch(e) {
@@ -435,3 +453,20 @@ function restartServer() {
   waitForRestart();
 }
 
+// legacy functions for CouchDB < 1.2.0
+// we keep them to make sure we keep BC
+CouchDB.user_prefix = "org.couchdb.user:";
+
+CouchDB.prepareUserDoc = function(user_doc, new_password) {
+  user_doc._id = user_doc._id || CouchDB.user_prefix + user_doc.name;
+  if (new_password) {
+    // handle the password crypto
+    user_doc.salt = CouchDB.newUuids(1)[0];
+    user_doc.password_sha = hex_sha1(new_password + user_doc.salt);
+  }
+  user_doc.type = "user";
+  if (!user_doc.roles) {
+    user_doc.roles = [];
+  }
+  return user_doc;
+};
