@@ -190,26 +190,34 @@
       }
 
       this.databaseSecurity = function() {
+        function namesAndRoles(r, key) {
+          var names = [];
+          var roles = [];
+          if (r && typeof r[key + "s"] === "object") {
+            if ($.isArray(r[key + "s"]["names"])) {
+              names = r[key + "s"]["names"];
+            }
+            if ($.isArray(r[key + "s"]["roles"])) {
+              roles = r[key + "s"]["roles"];
+            }
+          }
+          return {names : names, roles: roles};
+        };
+
         $.showDialog("dialog/_database_security.html", {
           load : function(d) {
             db.getDbProperty("_security", {
               success: function(r) {
-                ["admin", "reader"].forEach(function(key) {
-                  var names = [];
-                  var roles = [];
-
-                  if (r && typeof r[key + "s"] === "object") {
-                    if ($.isArray(r[key + "s"]["names"])) {
-                      names = r[key + "s"]["names"];
-                    }
-                    if ($.isArray(r[key + "s"]["roles"])) {
-                      roles = r[key + "s"]["roles"];
-                    }
-                  }
-
-                  $("input[name=" + key + "_names]", d).val(JSON.stringify(names));
-                  $("input[name=" + key + "_roles]", d).val(JSON.stringify(roles));
-                });
+                var admins = namesAndRoles(r, "admin")
+                  , members = namesAndRoles(r, "member");
+                if (members.names.length + members.roles.length == 0) {
+                  // backwards compatibility with readers for 1.x
+                  members = namesAndRoles(r, "reader");
+                }
+                $("input[name=admin_names]", d).val(JSON.stringify(admins.names));
+                $("input[name=admin_roles]", d).val(JSON.stringify(admins.roles));
+                $("input[name=member_names]", d).val(JSON.stringify(members.names));
+                $("input[name=member_roles]", d).val(JSON.stringify(members.roles));
               }
             });
           },
@@ -221,13 +229,13 @@
                 names: [],
                 roles: []
               },
-              readers: {
+              members: {
                 names: [],
                 roles: []
               }
             };
 
-            ["admin", "reader"].forEach(function(key) {
+            ["admin", "member"].forEach(function(key) {
               var names, roles;
 
               try {
@@ -529,6 +537,9 @@
                     location.href = "database.html?" + encodeURIComponent(dbName) +
                       "/" + $.couch.encodeDocId(doc._id) +
                       "/_view/" + encodeURIComponent(data.name);
+                  },
+                  error: function(status, e, reason) {
+                    alert(reason);
                   }
                 });
               }
@@ -569,8 +580,12 @@
             db.saveDoc(doc, {
               success: function(resp) {
                 page.isDirty = false;
+                page.storedViewCode = viewDef;
                 $("#viewcode button.revert, #viewcode button.save")
                   .attr("disabled", "disabled");
+              },
+              error: function(status, e, reason) {
+                alert(reason);
               }
             });
           }
@@ -1275,8 +1290,7 @@
           return false;
         }).prependTo($("a", li));
       }
-    },
-
+    }
   });
 
   function encodeAttachment(name) {
