@@ -70,6 +70,14 @@ couchTests.etags_views = function(debug) {
   xhr = CouchDB.request("GET", "/test_suite_db/_design/etags/_view/basicView");
   var etag1 = xhr.getResponseHeader("etag");
   T(etag1 == etag);
+
+  // verify ETag always changes for include_docs=true on update
+  xhr = CouchDB.request("GET", "/test_suite_db/_design/etags/_view/basicView?include_docs=true");
+  var etag1 = xhr.getResponseHeader("etag");
+  T(db.save({"_id":"doc2", "foo":"bar"}).ok);
+  xhr = CouchDB.request("GET", "/test_suite_db/_design/etags/_view/basicView?include_docs=true");
+  var etag2 = xhr.getResponseHeader("etag");
+  T(etag1 != etag2);
  
   // Verify that purges affect etags
   xhr = CouchDB.request("GET", "/test_suite_db/_design/etags/_view/fooView");
@@ -104,7 +112,7 @@ couchTests.etags_views = function(debug) {
   restartServer();
   xhr = CouchDB.request("GET", "/test_suite_db/_design/etags/_view/basicView");
   var etag2 = xhr.getResponseHeader("etag");
-  T(etag1 == etag2);
+  TEquals(etag2, etag1, "etag should be the same after restart 1");
 
   // reduce view
   xhr = CouchDB.request("GET", "/test_suite_db/_design/etags/_view/withReduce");
@@ -147,7 +155,7 @@ couchTests.etags_views = function(debug) {
   restartServer();
   xhr = CouchDB.request("GET", "/test_suite_db/_design/etags/_view/withReduce");
   var etag2 = xhr.getResponseHeader("etag");
-  T(etag1 == etag2);
+  TEquals(etag2, etag1, "etag should be the same after restart 2");
 
   // confirm ETag changes with different POST bodies
   xhr = CouchDB.request("POST", "/test_suite_db/_design/etags/_view/basicView",
