@@ -30,7 +30,8 @@
 -record(state, {
     max_cache_size = 0,
     cache_size = 0,
-    db_notifier = nil
+    db_notifier = nil,
+    db_mon_ref = nil
 }).
 
 
@@ -103,19 +104,13 @@ init(_) ->
     ?STATE = ets:new(?STATE, [set, protected, named_table]),
     ?BY_USER = ets:new(?BY_USER, [set, protected, named_table]),
     ?BY_ATIME = ets:new(?BY_ATIME, [ordered_set, private, named_table]),
-    AuthDbName = couch_config:get("couch_httpd_auth", "authentication_db"),
-    true = ets:insert(?STATE, {auth_db_name, ?l2b(AuthDbName)}),
-    true = ets:insert(?STATE, {auth_db, open_auth_db()}),
     process_flag(trap_exit, true),
     ok = couch_config:register(
         fun("couch_httpd_auth", "auth_cache_size", SizeList) ->
             Size = list_to_integer(SizeList),
-            ok = gen_server:call(?MODULE, {new_max_cache_size, Size}, infinity)
-        end
-    ),
-    ok = couch_config:register(
-        fun("couch_httpd_auth", "authentication_db", DbName) ->
-            ok = gen_server:call(?MODULE, {new_auth_db, ?l2b(DbName)}, infinity)
+            ok = gen_server:call(?MODULE, {new_max_cache_size, Size}, infinity);
+        ("couch_httpd_auth", "authentication_db", _DbName) ->
+            ok = gen_server:call(?MODULE, reinit_cache, infinity)
         end
     ),
     {ok, Notifier} = couch_db_update_notifier:start_link(fun handle_db_event/1),
@@ -125,7 +120,7 @@ init(_) ->
             couch_config:get("couch_httpd_auth", "auth_cache_size", "50")
         )
     },
-    {ok, State}.
+    {ok, reinit_cache(State)}.
 
 
 handle_db_event({Event, DbName}) ->
@@ -133,8 +128,7 @@ handle_db_event({Event, DbName}) ->
     case DbName =:= AuthDbName of
     true ->
         case Event of
-        deleted -> gen_server:call(?MODULE, auth_db_deleted, infinity);
-        created -> gen_server:call(?MODULE, auth_db_created, infinity);
+        created -> gen_server:call(?MODULE, reinit_cache, infinity);
         compacted -> gen_server:call(?MODULE, auth_db_compacted, infinity);
         _Else   -> ok
         end;
@@ -143,21 +137,10 @@ handle_db_event({Event, DbName}) ->
     end.
 
 
-handle_call({new_auth_db, AuthDbName}, _From, State) ->
-    NewState = clear_cache(State),
-    true = ets:insert(?STATE, {auth_db_name, AuthDbName}),
-    true = ets:insert(?STATE, {auth_db, open_auth_db()}),
-    {reply, ok, NewState};
-
-handle_call(auth_db_deleted, _From, State) ->
-    NewState = clear_cache(State),
-    true = ets:insert(?STATE, {auth_db, nil}),
-    {reply, ok, NewState};
-
-handle_call(auth_db_created, _From, State) ->
-    NewState = clear_cache(State),
-    true = ets:insert(?STATE, {auth_db, open_auth_db()}),
-    {reply, ok, NewState};
+handle_call(reinit_cache, _From, State) ->
+    catch erlang:demonitor(State#state.db_mon_ref, [flush]),
+    exec_if_auth_db(fun(AuthDb) -> catch couch_db:close(AuthDb) end),
+    {reply, ok, reinit_cache(State)};
 
 handle_call(auth_db_compacted, _From, State) ->
     exec_if_auth_db(
@@ -167,26 +150,13 @@ handle_call(auth_db_compacted, _From, State) ->
     ),
     {reply, ok, State};
 
+handle_call({new_max_cache_size, NewSize},
+        _From, #state{cache_size = Size} = State) when NewSize >= Size ->
+    {reply, ok, State#state{max_cache_size = NewSize}};
+
 handle_call({new_max_cache_size, NewSize}, _From, State) ->
-    case NewSize >= State#state.cache_size of
-    true ->
-        ok;
-    false ->
-        lists:foreach(
-            fun(_) ->
-                LruTime = ets:last(?BY_ATIME),
-                [{LruTime, UserName}] = ets:lookup(?BY_ATIME, LruTime),
-                true = ets:delete(?BY_ATIME, LruTime),
-                true = ets:delete(?BY_USER, UserName)
-            end,
-            lists:seq(1, State#state.cache_size - NewSize)
-        )
-    end,
-    NewState = State#state{
-        max_cache_size = NewSize,
-        cache_size = lists:min([NewSize, State#state.cache_size])
-    },
-    {reply, ok, NewState};
+    free_mru_cache_entries(State#state.cache_size - NewSize),
+    {reply, ok, State#state{max_cache_size = NewSize, cache_size = NewSize}};
 
 handle_call({fetch, UserName}, _From, State) ->
     {Credentials, NewState} = case ets:lookup(?BY_USER, UserName) of
@@ -217,8 +187,8 @@ handle_cast({cache_hit, UserName}, State) ->
     {noreply, State}.
 
 
-handle_info(_Msg, State) ->
-    {noreply, State}.
+handle_info({'DOWN', Ref, _, _, _Reason}, #state{db_mon_ref = Ref} = State) ->
+    {noreply, reinit_cache(State)}.
 
 
 terminate(_Reason, #state{db_notifier = Notifier}) ->
@@ -240,6 +210,17 @@ clear_cache(State) ->
     State#state{cache_size = 0}.
 
 
+reinit_cache(State) ->
+    NewState = clear_cache(State),
+    AuthDbName = ?l2b(couch_config:get("couch_httpd_auth", "authentication_db")),
+    true = ets:insert(?STATE, {auth_db_name, AuthDbName}),
+    AuthDb = open_auth_db(),
+    true = ets:insert(?STATE, {auth_db, AuthDb}),
+    NewState#state{db_mon_ref = couch_db:monitor(AuthDb)}.
+
+
+add_cache_entry(_, _, _, #state{max_cache_size = 0} = State) ->
+    State;
 add_cache_entry(UserName, Credentials, ATime, State) ->
     case State#state.cache_size >= State#state.max_cache_size of
     true ->
@@ -251,16 +232,17 @@ add_cache_entry(UserName, Credentials, ATime, State) ->
     true = ets:insert(?BY_USER, {UserName, {Credentials, ATime}}),
     State#state{cache_size = couch_util:get_value(size, ets:info(?BY_USER))}.
 
+free_mru_cache_entries(0) ->
+    ok;
+free_mru_cache_entries(N) when N > 0 ->
+    free_mru_cache_entry(),
+    free_mru_cache_entries(N - 1).
 
 free_mru_cache_entry() ->
-    case ets:last(?BY_ATIME) of
-    '$end_of_table' ->
-        ok;  % empty cache
-    LruTime ->
-        [{LruTime, UserName}] = ets:lookup(?BY_ATIME, LruTime),
-        true = ets:delete(?BY_ATIME, LruTime),
-        true = ets:delete(?BY_USER, UserName)
-    end.
+    MruTime = ets:last(?BY_ATIME),
+    [{MruTime, UserName}] = ets:lookup(?BY_ATIME, MruTime),
+    true = ets:delete(?BY_ATIME, MruTime),
+    true = ets:delete(?BY_USER, UserName).
 
 
 cache_hit(UserName, Credentials, ATime) ->
@@ -405,8 +387,17 @@ ensure_auth_ddoc_exists(Db, DDocId) ->
     {not_found, _Reason} ->
         {ok, AuthDesign} = auth_design_doc(DDocId),
         {ok, _Rev} = couch_db:update_doc(Db, AuthDesign, []);
-    _ ->
-        ok
+    {ok, Doc} ->
+        {Props} = couch_doc:to_json_obj(Doc, []),
+        case couch_util:get_value(<<"validate_doc_update">>, Props, []) of
+            ?AUTH_DB_DOC_VALIDATE_FUNCTION ->
+                ok;
+            _ ->
+                Props1 = lists:keyreplace(<<"validate_doc_update">>, 1, Props,
+                    {<<"validate_doc_update">>,
+                    ?AUTH_DB_DOC_VALIDATE_FUNCTION}),
+                couch_db:update_doc(Db, couch_doc:from_json_obj({Props1}), [])
+        end
     end,
     ok.
 
