@@ -13,7 +13,8 @@
 -module(couch_httpd).
 -include("couch_db.hrl").
 
--export([start_link/0, start_link/1, stop/0, handle_request/5]).
+-export([start_link/0, start_link/1, stop/0, config_change/2, 
+        handle_request/5]).
 
 -export([header_value/2,header_value/3,qs_value/2,qs_value/3,qs/1,qs_json_value/3]).
 -export([path/1,absolute_uri/2,body_length/1]).
@@ -39,16 +40,54 @@ start_link(https) ->
     CertFile = couch_config:get("ssl", "cert_file", nil),
     KeyFile = couch_config:get("ssl", "key_file", nil),
     Options = case CertFile /= nil andalso KeyFile /= nil of
-                  true ->
-                      [{port, Port},
-                       {ssl, true},
-                       {ssl_opts, [
-                             {certfile, CertFile},
-                             {keyfile, KeyFile}]}];
-                  false ->
-                      io:format("SSL enabled but PEM certificates are missing.", []),
-                      throw({error, missing_certs})
-              end,
+        true ->
+            SslOpts = [{certfile, CertFile}, {keyfile, KeyFile}],
+
+            %% set password if one is needed for the cert
+            SslOpts1 = case couch_config:get("ssl", "password", nil) of
+                nil -> SslOpts;
+                Password ->
+                    SslOpts ++ [{password, Password}]
+            end,
+            % do we verify certificates ?
+            FinalSslOpts = case couch_config:get("ssl",
+                    "verify_ssl_certificates", "false") of
+                "false" -> SslOpts1;
+                "true" ->
+                    case couch_config:get("ssl",
+                            "cacert_file", nil) of
+                        nil ->
+                            io:format("Verify SSL certificate "
+                                ++"enabled but file containing "
+                                ++"PEM encoded CA certificates is "
+                                ++"missing", []),
+                            throw({error, missing_cacerts});
+                        CaCertFile ->
+                            Depth = list_to_integer(couch_config:get("ssl",
+                                    "ssl_certificate_max_depth",
+                                    "1")),
+                            FinalOpts = [
+                                {cacertfile, CaCertFile},
+                                {depth, Depth},
+                                {verify, verify_peer}],
+                            % allows custom verify fun.
+                            case couch_config:get("ssl",
+                                    "verify_fun", nil) of
+                                nil -> FinalOpts;
+                                SpecStr ->
+                                    FinalOpts 
+                                    ++ [{verify_fun, make_arity_3_fun(SpecStr)}]
+                            end
+                    end
+            end,
+            
+            [{port, Port},
+                {ssl, true},
+                {ssl_opts, FinalSslOpts}];
+        false ->
+            io:format("SSL enabled but PEM certificates are missing.", []),
+            throw({error, missing_certs})
+    end,
     start_link(https, Options).
 start_link(Name, Options) ->
     % read config and register for configuration changes
@@ -84,6 +123,9 @@ start_link(Name, Options) ->
         couch_config:get("httpd", "server_options", "[]")),
     {ok, SocketOptions} = couch_util:parse_term(
         couch_config:get("httpd", "socket_options", "[]")),
+
+    set_auth_handlers(),
+
     Loop = fun(Req)->
         case SocketOptions of
         [] ->
@@ -96,42 +138,55 @@ start_link(Name, Options) ->
         ])
     end,
 
-    % and off we go
+    % set mochiweb options
+    FinalOptions = lists:append([Options, ServerOptions, [
+            {loop, Loop},
+            {name, Name},
+            {ip, BindAddress}]]),
 
-    {ok, Pid} = case mochiweb_http:start(Options ++ ServerOptions ++ [
-        {loop, Loop},
-        {name, Name},
-        {ip, BindAddress}
-    ]) of
-    {ok, MochiPid} -> {ok, MochiPid};
-    {error, Reason} ->
-        io:format("Failure to start Mochiweb: ~s~n",[Reason]),
-        throw({error, Reason})
+    % launch mochiweb
+    {ok, Pid} = case mochiweb_http:start(FinalOptions) of
+        {ok, MochiPid} -> 
+            {ok, MochiPid};
+        {error, Reason} ->
+            io:format("Failure to start Mochiweb: ~s~n",[Reason]),
+            throw({error, Reason})
     end,
 
-    ok = couch_config:register(
-        fun("httpd", "bind_address") ->
-            ?MODULE:stop();
-        ("httpd", "port") ->
-            ?MODULE:stop();
-        ("httpd", "default_handler") ->
-            ?MODULE:stop();
-        ("httpd", "server_options") ->
-            ?MODULE:stop();
-        ("httpd", "socket_options") ->
-            ?MODULE:stop();
-        ("httpd_global_handlers", _) ->
-            ?MODULE:stop();
-        ("httpd_db_handlers", _) ->
-            ?MODULE:stop();
-        ("vhosts", _) ->
-            ?MODULE:stop();
-        ("ssl", _) ->
-            ?MODULE:stop()
-        end, Pid),
-
+    ok = couch_config:register(fun ?MODULE:config_change/2, Pid),
     {ok, Pid}.
 
+
+stop() ->
+    mochiweb_http:stop(couch_httpd),
+    mochiweb_http:stop(https).
+
+config_change("httpd", "bind_address") ->
+    ?MODULE:stop();
+config_change("httpd", "port") ->
+    ?MODULE:stop();
+config_change("httpd", "default_handler") ->
+    ?MODULE:stop();
+config_change("httpd", "server_options") ->
+    ?MODULE:stop();
+config_change("httpd", "socket_options") ->
+    ?MODULE:stop();
+config_change("httpd", "authentication_handlers") ->
+    set_auth_handlers();
+config_change("httpd_global_handlers", _) ->
+    ?MODULE:stop();
+config_change("httpd_db_handlers", _) ->
+    ?MODULE:stop();
+config_change("ssl", _) ->
+    ?MODULE:stop().
+
+set_auth_handlers() ->
+    AuthenticationSrcs = make_fun_spec_strs(
+        couch_config:get("httpd", "authentication_handlers", "")),
+    AuthHandlers = lists:map(
+        fun(A) -> {make_arity_1_fun(A), ?l2b(A)} end, AuthenticationSrcs),
+    ok = application:set_env(couch, auth_handlers, AuthHandlers).
+
 % SpecStr is a string like "{my_module, my_fun}"
 %  or "{my_module, my_fun, <<"my_arg">>}"
 make_arity_1_fun(SpecStr) ->
@@ -162,22 +217,17 @@ make_arity_3_fun(SpecStr) ->
 make_fun_spec_strs(SpecStr) ->
     re:split(SpecStr, "(?<=})\\s*,\\s*(?={)", [{return, list}]).
 
-stop() ->
-    mochiweb_http:stop(?MODULE).
-
-
 handle_request(MochiReq, DefaultFun, UrlHandlers, DbUrlHandlers, 
     DesignUrlHandlers) ->
 
-    MochiReq1 = couch_httpd_vhost:match_vhost(MochiReq),
+    MochiReq1 = couch_httpd_vhost:dispatch_host(MochiReq),
+    
     handle_request_int(MochiReq1, DefaultFun,
                 UrlHandlers, DbUrlHandlers, DesignUrlHandlers).
 
 handle_request_int(MochiReq, DefaultFun,
             UrlHandlers, DbUrlHandlers, DesignUrlHandlers) ->
     Begin = now(),
-    AuthenticationSrcs = make_fun_spec_strs(
-            couch_config:get("httpd", "authentication_handlers")),
     % for the path, use the raw path with the query string and fragment
     % removed, but URL quoting left intact
     RawUri = MochiReq:get(raw_path),
@@ -187,7 +237,11 @@ handle_request_int(MochiReq, DefaultFun,
 
     % get requested path
     RequestedPath = case MochiReq:get_header_value("x-couchdb-vhost-path") of
-        undefined -> RawUri;
+        undefined ->
+            case MochiReq:get_header_value("x-couchdb-requested-path") of
+                undefined -> RawUri;
+                R -> R
+            end;
         P -> P
     end,
     
@@ -235,7 +289,7 @@ handle_request_int(MochiReq, DefaultFun,
 
     % alias HEAD to GET as mochiweb takes care of stripping the body
     Method = case Method2 of
-            'HEAD' -> 'GET';
+        'HEAD' -> 'GET';
         Other -> Other
     end,
 
@@ -243,21 +297,22 @@ handle_request_int(MochiReq, DefaultFun,
         mochi_req = MochiReq,
         peer = MochiReq:get(peer),
         method = Method,
-        requested_path_parts = [list_to_binary(couch_httpd:unquote(Part))
-                || Part <- string:tokens(RequestedPath, "/")],
-        path_parts = [list_to_binary(couch_httpd:unquote(Part))
-                || Part <- string:tokens(Path, "/")],
+        requested_path_parts =
+            [?l2b(unquote(Part)) || Part <- string:tokens(RequestedPath, "/")],
+        path_parts = [?l2b(unquote(Part)) || Part <- string:tokens(Path, "/")],
         db_url_handlers = DbUrlHandlers,
         design_url_handlers = DesignUrlHandlers,
         default_fun = DefaultFun,
-        url_handlers = UrlHandlers
+        url_handlers = UrlHandlers,
+        user_ctx = erlang:erase(pre_rewrite_user_ctx)
     },
 
     HandlerFun = couch_util:dict_find(HandlerKey, UrlHandlers, DefaultFun),
+    {ok, AuthHandlers} = application:get_env(couch, auth_handlers),
 
     {ok, Resp} =
     try
-        case authenticate_request(HttpReq, AuthenticationSrcs) of
+        case authenticate_request(HttpReq, AuthHandlers) of
         #httpd{} = Req ->
             HandlerFun(Req);
         Response ->
@@ -269,7 +324,7 @@ handle_request_int(MochiReq, DefaultFun,
         throw:{invalid_json, S} ->
             ?LOG_ERROR("attempted upload of invalid JSON (set log_level to debug to log it)", []),
             ?LOG_DEBUG("Invalid JSON: ~p",[S]),
-            send_error(HttpReq, {bad_request, io_lib:format("invalid UTF-8 JSON: ~p",[S])});
+            send_error(HttpReq, {bad_request, invalid_json});
         throw:unacceptable_encoding ->
             ?LOG_ERROR("unsupported encoding method for the response", []),
             send_error(HttpReq, {not_acceptable, "unsupported encoding"});
@@ -278,21 +333,30 @@ handle_request_int(MochiReq, DefaultFun,
             send_error(HttpReq, bad_request);
         exit:normal ->
             exit(normal);
+        exit:snappy_nif_not_loaded ->
+            ErrorReason = "To access the database or view index, Apache CouchDB"
+                " must be built with Erlang OTP R13B04 or higher.",
+            ?LOG_ERROR("~s", [ErrorReason]),
+            send_error(HttpReq, {bad_otp_release, ErrorReason});
         throw:Error ->
+            Stack = erlang:get_stacktrace(),
             ?LOG_DEBUG("Minor error in HTTP request: ~p",[Error]),
-            ?LOG_DEBUG("Stacktrace: ~p",[erlang:get_stacktrace()]),
+            ?LOG_DEBUG("Stacktrace: ~p",[Stack]),
             send_error(HttpReq, Error);
         error:badarg ->
+            Stack = erlang:get_stacktrace(),
             ?LOG_ERROR("Badarg error in HTTP request",[]),
-            ?LOG_INFO("Stacktrace: ~p",[erlang:get_stacktrace()]),
+            ?LOG_INFO("Stacktrace: ~p",[Stack]),
             send_error(HttpReq, badarg);
         error:function_clause ->
+            Stack = erlang:get_stacktrace(),
             ?LOG_ERROR("function_clause error in HTTP request",[]),
-            ?LOG_INFO("Stacktrace: ~p",[erlang:get_stacktrace()]),
+            ?LOG_INFO("Stacktrace: ~p",[Stack]),
             send_error(HttpReq, function_clause);
         Tag:Error ->
+            Stack = erlang:get_stacktrace(),
             ?LOG_ERROR("Uncaught error in HTTP request: ~p",[{Tag, Error}]),
-            ?LOG_INFO("Stacktrace: ~p",[erlang:get_stacktrace()]),
+            ?LOG_INFO("Stacktrace: ~p",[Stack]),
             send_error(HttpReq, Error)
     end,
     RequestTime = round(timer:now_diff(now(), Begin)/1000),
@@ -303,7 +367,7 @@ handle_request_int(MochiReq, DefaultFun,
 % Try authentication handlers in order until one sets a user_ctx
 % the auth funs also have the option of returning a response
 % move this to couch_httpd_auth?
-authenticate_request(#httpd{user_ctx=#user_ctx{}} = Req, _AuthSrcs) ->
+authenticate_request(#httpd{user_ctx=#user_ctx{}} = Req, _AuthHandlers) ->
     Req;
 authenticate_request(#httpd{} = Req, []) ->
     case couch_config:get("couch_httpd_auth", "require_valid_user", "false") of
@@ -312,14 +376,13 @@ authenticate_request(#httpd{} = Req, []) ->
     "false" ->
         Req#httpd{user_ctx=#user_ctx{}}
     end;
-authenticate_request(#httpd{} = Req, [AuthSrc|Rest]) ->
-    AuthFun = make_arity_1_fun(AuthSrc),
+authenticate_request(#httpd{} = Req, [{AuthFun, AuthSrc} | RestAuthHandlers]) ->
     R = case AuthFun(Req) of
         #httpd{user_ctx=#user_ctx{}=UserCtx}=Req2 ->
-            Req2#httpd{user_ctx=UserCtx#user_ctx{handler=?l2b(AuthSrc)}};
+            Req2#httpd{user_ctx=UserCtx#user_ctx{handler=AuthSrc}};
         Else -> Else
     end,
-    authenticate_request(R, Rest);
+    authenticate_request(R, RestAuthHandlers);
 authenticate_request(Response, _AuthSrcs) ->
     Response.
 
@@ -341,12 +404,11 @@ validate_referer(Req) ->
     end.
 
 validate_ctype(Req, Ctype) ->
-    case couch_httpd:header_value(Req, "Content-Type") of
+    case header_value(Req, "Content-Type") of
     undefined ->
         throw({bad_ctype, "Content-Type must be "++Ctype});
     ReqCtype ->
-        % ?LOG_ERROR("Ctype ~p ReqCtype ~p",[Ctype,ReqCtype]),
-        case re:split(ReqCtype, ";", [{return, list}]) of
+        case string:tokens(ReqCtype, ";") of
         [Ctype] -> ok;
         [Ctype, _Rest] -> ok;
         _Else ->
@@ -385,6 +447,7 @@ serve_file(Req, RelativePath, DocumentRoot) ->
     serve_file(Req, RelativePath, DocumentRoot, []).
 
 serve_file(#httpd{mochi_req=MochiReq}=Req, RelativePath, DocumentRoot, ExtraHeaders) ->
+    log_request(Req, 200),
     {ok, MochiReq:serve_file(RelativePath, DocumentRoot,
         server_header() ++ couch_httpd_auth:cookie_auth_header(Req, []) ++ ExtraHeaders)}.
 
@@ -469,16 +532,24 @@ body_length(Req) ->
         Unknown -> {unknown_transfer_encoding, Unknown}
     end.
 
-body(#httpd{mochi_req=MochiReq, req_body=ReqBody}) ->
-    case ReqBody of
+body(#httpd{mochi_req=MochiReq, req_body=undefined} = Req) ->
+    case body_length(Req) of
         undefined ->
-            % Maximum size of document PUT request body (4GB)
             MaxSize = list_to_integer(
                 couch_config:get("couchdb", "max_document_size", "4294967296")),
             MochiReq:recv_body(MaxSize);
-        _Else ->
-            ReqBody
-    end.
+        chunked ->
+            ChunkFun = fun({0, _Footers}, Acc) ->
+                lists:reverse(Acc);
+            ({_Len, Chunk}, Acc) ->
+                [Chunk | Acc]
+            end,
+            recv_chunked(Req, 8192, ChunkFun, []);
+        Len ->
+            MochiReq:recv_body(Len)
+    end;
+body(#httpd{req_body=ReqBody}) ->
+    ReqBody.
 
 json_body(Httpd) ->
     ?JSON_DECODE(body(Httpd)).
@@ -497,21 +568,21 @@ doc_etag(#doc{revs={Start, [DiskRev|_]}}) ->
 
 make_etag(Term) ->
     <<SigInt:128/integer>> = couch_util:md5(term_to_binary(Term)),
-    list_to_binary("\"" ++ lists:flatten(io_lib:format("~.36B",[SigInt])) ++ "\"").
+    iolist_to_binary([$", io_lib:format("~.36B", [SigInt]), $"]).
 
 etag_match(Req, CurrentEtag) when is_binary(CurrentEtag) ->
     etag_match(Req, binary_to_list(CurrentEtag));
 
 etag_match(Req, CurrentEtag) ->
     EtagsToMatch = string:tokens(
-        couch_httpd:header_value(Req, "If-None-Match", ""), ", "),
+        header_value(Req, "If-None-Match", ""), ", "),
     lists:member(CurrentEtag, EtagsToMatch).
 
 etag_respond(Req, CurrentEtag, RespFun) ->
     case etag_match(Req, CurrentEtag) of
     true ->
         % the client has this in their cache.
-        couch_httpd:send_response(Req, 304, [{"Etag", CurrentEtag}], <<>>);
+        send_response(Req, 304, [{"ETag", CurrentEtag}], <<>>);
     false ->
         % Run the function.
         RespFun()
@@ -526,11 +597,11 @@ verify_is_server_admin(#user_ctx{roles=Roles}) ->
     end.
 
 log_request(#httpd{mochi_req=MochiReq,peer=Peer}, Code) ->
-    ?LOG_INFO("~s - - ~p ~s ~B", [
+    ?LOG_INFO("~s - - ~s ~s ~B", [
         Peer,
-        couch_util:to_existing_atom(MochiReq:get(method)),
+        MochiReq:get(method),
         MochiReq:get(raw_path),
-        couch_util:to_integer(Code)
+        Code
     ]).
 
 
@@ -619,25 +690,25 @@ send_json(Req, Code, Value) ->
     send_json(Req, Code, [], Value).
 
 send_json(Req, Code, Headers, Value) ->
+    initialize_jsonp(Req),
     DefaultHeaders = [
         {"Content-Type", negotiate_content_type(Req)},
         {"Cache-Control", "must-revalidate"}
     ],
-    Body = [start_jsonp(Req), ?JSON_ENCODE(Value), end_jsonp(), $\n],
+    Body = [start_jsonp(), ?JSON_ENCODE(Value), end_jsonp(), $\n],
     send_response(Req, Code, DefaultHeaders ++ Headers, Body).
 
 start_json_response(Req, Code) ->
     start_json_response(Req, Code, []).
 
 start_json_response(Req, Code, Headers) ->
+    initialize_jsonp(Req),
     DefaultHeaders = [
         {"Content-Type", negotiate_content_type(Req)},
         {"Cache-Control", "must-revalidate"}
     ],
-    start_jsonp(Req), % Validate before starting chunked.
-    %start_chunked_response(Req, Code, DefaultHeaders ++ Headers).
     {ok, Resp} = start_chunked_response(Req, Code, DefaultHeaders ++ Headers),
-    case start_jsonp(Req) of
+    case start_jsonp() of
         [] -> ok;
         Start -> send_chunk(Resp, Start)
     end,
@@ -647,7 +718,7 @@ end_json_response(Resp) ->
     send_chunk(Resp, end_jsonp() ++ [$\n]),
     last_chunk(Resp).
 
-start_jsonp(Req) ->
+initialize_jsonp(Req) ->
     case get(jsonp) of
         undefined -> put(jsonp, qs_value(Req, "callback", no_jsonp));
         _ -> ok
@@ -660,14 +731,9 @@ start_jsonp(Req) ->
                 % make sure jsonp is configured on (default off)
                 case couch_config:get("httpd", "allow_jsonp", "false") of
                 "true" ->
-                    validate_callback(CallBack),
-                    CallBack ++ "(";
+                    validate_callback(CallBack);
                 _Else ->
-                    % this could throw an error message, but instead we just ignore the
-                    % jsonp parameter
-                    % throw({bad_request, <<"JSONP must be configured before using.">>})
-                    put(jsonp, no_jsonp),
-                    []
+                    put(jsonp, no_jsonp)
                 end
             catch
                 Error ->
@@ -676,14 +742,19 @@ start_jsonp(Req) ->
             end
     end.
 
+start_jsonp() ->
+    case get(jsonp) of
+        no_jsonp -> [];
+        [] -> [];
+        CallBack -> CallBack ++ "("
+    end.
+
 end_jsonp() ->
-    Resp = case get(jsonp) of
+    case erlang:erase(jsonp) of
         no_jsonp -> [];
         [] -> [];
         _ -> ");"
-    end,
-    put(jsonp, undefined),
-    Resp.
+    end.
 
 validate_callback(CallBack) when is_binary(CallBack) ->
     validate_callback(binary_to_list(CallBack));
@@ -741,8 +812,10 @@ error_info({error, illegal_database_name}) ->
 error_info({missing_stub, Reason}) ->
     {412, <<"missing_stub">>, Reason};
 error_info({Error, Reason}) ->
+    ?LOG_ERROR("Uncaught server error: ~p", [{Error, Reason}]),
     {500, couch_util:to_binary(Error), couch_util:to_binary(Reason)};
 error_info(Error) ->
+    ?LOG_ERROR("Uncaught server error: ~p", [Error]),
     {500, <<"unknown_error">>, couch_util:to_binary(Error)}.
 
 error_headers(#httpd{mochi_req=MochiReq}=Req, Code, ErrorStr, ReasonStr) ->
@@ -833,10 +906,16 @@ send_chunked_error(Resp, Error) ->
     last_chunk(Resp).
 
 send_redirect(Req, Path) ->
-     Headers = [{"Location", couch_httpd:absolute_uri(Req, Path)}],
-     send_response(Req, 301, Headers, <<>>).
+     send_response(Req, 301, [{"Location", absolute_uri(Req, Path)}], <<>>).
 
-negotiate_content_type(#httpd{mochi_req=MochiReq}) ->
+negotiate_content_type(Req) ->
+    case get(jsonp) of
+        no_jsonp -> negotiate_content_type1(Req);
+        [] -> negotiate_content_type1(Req);
+        _Callback -> "text/javascript"
+    end.
+
+negotiate_content_type1(#httpd{mochi_req=MochiReq}) ->
     %% Determine the appropriate Content-Type header for a JSON response
     %% depending on the Accept header in the request. A request that explicitly
     %% lists the correct JSON MIME type will get that type, otherwise the
@@ -847,7 +926,7 @@ negotiate_content_type(#httpd{mochi_req=MochiReq}) ->
     end,
     case lists:member("application/json", AcceptedTypes) of
         true  -> "application/json";
-        false -> "text/plain;charset=utf-8"
+        false -> "text/plain; charset=utf-8"
     end.
 
 server_header() ->
@@ -866,13 +945,13 @@ parse_multipart_request(ContentType, DataFun, Callback) ->
             data_fun=DataFun,
             callback=Callback},
     {Mp2, _NilCallback} = read_until(Mp, <<"--", Boundary0/binary>>,
-        fun(Next)-> nil_callback(Next) end),
+        fun nil_callback/1),
     #mp{buffer=Buffer, data_fun=DataFun2, callback=Callback2} =
             parse_part_header(Mp2),
     {Buffer, DataFun2, Callback2}.
 
 nil_callback(_Data)->
-    fun(Next) -> nil_callback(Next) end.
+    fun nil_callback/1.
 
 get_boundary({"multipart/" ++ _, Opts}) ->
     case couch_util:get_value("boundary", Opts) of
