@@ -20,7 +20,7 @@
 
 -include("couch_db.hrl").
 
--define(PORT_OPTIONS, [stream, {line, 1024}, binary, exit_status, hide]).
+-define(PORT_OPTIONS, [stream, {line, 4096}, binary, exit_status, hide]).
 
 -record(os_proc,
     {command,
@@ -42,7 +42,7 @@ stop(Pid) ->
 
 % Read/Write API
 set_timeout(Pid, TimeOut) when is_integer(TimeOut) ->
-    ok = gen_server:call(Pid, {set_timeout, TimeOut}).
+    ok = gen_server:call(Pid, {set_timeout, TimeOut}, infinity).
 
 % Used by couch_db_update_notifier.erl
 send(Pid, Data) ->
@@ -60,14 +60,22 @@ prompt(Pid, Data) ->
 % Utility functions for reading and writing
 % in custom functions
 writeline(OsProc, Data) when is_record(OsProc, os_proc) ->
-    port_command(OsProc#os_proc.port, Data ++ "\n").
+    port_command(OsProc#os_proc.port, [Data, $\n]).
 
 readline(#os_proc{} = OsProc) ->
     readline(OsProc, []).
 readline(#os_proc{port = Port} = OsProc, Acc) ->
     receive
+    {Port, {data, {noeol, Data}}} when is_binary(Acc) ->
+        readline(OsProc, <<Acc/binary,Data/binary>>);
+    {Port, {data, {noeol, Data}}} when is_binary(Data) ->
+        readline(OsProc, Data);
     {Port, {data, {noeol, Data}}} ->
         readline(OsProc, [Data|Acc]);
+    {Port, {data, {eol, <<Data/binary>>}}} when is_binary(Acc) ->
+        [<<Acc/binary,Data/binary>>];
+    {Port, {data, {eol, Data}}} when is_binary(Data) ->
+        [Data];
     {Port, {data, {eol, Data}}} ->
         lists:reverse(Acc, Data);
     {Port, Err} ->
@@ -85,22 +93,51 @@ writejson(OsProc, Data) when is_record(OsProc, os_proc) ->
     true = writeline(OsProc, JsonData).
 
 readjson(OsProc) when is_record(OsProc, os_proc) ->
-    Line = readline(OsProc),
+    Line = iolist_to_binary(readline(OsProc)),
     ?LOG_DEBUG("OS Process ~p Output :: ~s", [OsProc#os_proc.port, Line]),
-    case ?JSON_DECODE(Line) of
-    [<<"log">>, Msg] when is_binary(Msg) ->
-        % we got a message to log. Log it and continue
-        ?LOG_INFO("OS Process ~p Log :: ~s", [OsProc#os_proc.port, Msg]),
-        readjson(OsProc);
-    [<<"error">>, Id, Reason] ->
-        throw({couch_util:to_existing_atom(Id),Reason});
-    [<<"fatal">>, Id, Reason] ->
-        ?LOG_INFO("OS Process ~p Fatal Error :: ~s ~p",[OsProc#os_proc.port, Id, Reason]),
-        throw({couch_util:to_existing_atom(Id),Reason});
-    Result ->
-        Result
+    try
+        % Don't actually parse the whole JSON. Just try to see if it's
+        % a command or a doc map/reduce/filter/show/list/update output.
+        % If it's a command then parse the whole JSON and execute the
+        % command, otherwise return the raw JSON line to the caller.
+        pick_command(Line)
+    catch
+    throw:abort ->
+        {json, Line};
+    throw:{cmd, _Cmd} ->
+        case ?JSON_DECODE(Line) of
+        [<<"log">>, Msg] when is_binary(Msg) ->
+            % we got a message to log. Log it and continue
+            ?LOG_INFO("OS Process ~p Log :: ~s", [OsProc#os_proc.port, Msg]),
+            readjson(OsProc);
+        [<<"error">>, Id, Reason] ->
+            throw({error, {couch_util:to_existing_atom(Id),Reason}});
+        [<<"fatal">>, Id, Reason] ->
+            ?LOG_INFO("OS Process ~p Fatal Error :: ~s ~p",
+                [OsProc#os_proc.port, Id, Reason]),
+            throw({couch_util:to_existing_atom(Id),Reason});
+        _Result ->
+            {json, Line}
+        end
     end.
 
+pick_command(Line) ->
+    json_stream_parse:events(Line, fun pick_command0/1).
+
+pick_command0(array_start) ->
+    fun pick_command1/1;
+pick_command0(_) ->
+    throw(abort).
+
+pick_command1(<<"log">> = Cmd) ->
+    throw({cmd, Cmd});
+pick_command1(<<"error">> = Cmd) ->
+    throw({cmd, Cmd});
+pick_command1(<<"fatal">> = Cmd) ->
+    throw({cmd, Cmd});
+pick_command1(_) ->
+    throw(abort).
+
 
 % gen_server API
 init([Command, Options, PortOptions]) ->
@@ -119,7 +156,7 @@ init([Command, Options, PortOptions]) ->
             % this ensure the real os process is killed when this process dies.
             erlang:monitor(process, Pid),
             receive _ -> ok end,
-            os:cmd(?b2l(KillCmd))
+            os:cmd(?b2l(iolist_to_binary(KillCmd)))
         end),
     OsProc =
     lists:foldl(fun(Opt, Proc) ->
@@ -148,8 +185,6 @@ handle_call({prompt, Data}, _From, OsProc) ->
     catch
         throw:{error, OsError} ->
             {reply, OsError, OsProc};
-        throw:{fatal, OsError} ->
-            {stop, normal, OsError, OsProc};
         throw:OtherError ->
             {stop, normal, OtherError, OsProc}
     end.
