@@ -76,9 +76,10 @@ get_temp_group(Db, Language, DesignOptions, MapSrc, RedSrc) ->
         get_temp_updater(couch_db:name(Db), Language, DesignOptions, MapSrc, RedSrc),
         couch_db:get_update_seq(Db)).
 
-get_group_info(Db, GroupId) ->
-    couch_view_group:request_group_info(
-        get_group_server(couch_db:name(Db), GroupId)).
+get_group_info(#db{name = DbName}, GroupId) ->
+    get_group_info(DbName, GroupId);
+get_group_info(DbName, GroupId) ->
+    couch_view_group:request_group_info(get_group_server(DbName, GroupId)).
 
 cleanup_index_files(Db) ->
     % load all ddocs
@@ -92,13 +93,18 @@ cleanup_index_files(Db) ->
 
     FileList = list_index_files(Db),
 
-    % regex that matches all ddocs
-    RegExp = "("++ string:join(Sigs, "|") ++")",
+    DeleteFiles =
+    if length(Sigs) =:= 0 ->
+        FileList;
+    true ->
+        % regex that matches all ddocs
+        RegExp = "("++ string:join(Sigs, "|") ++")",
+
+        % filter out the ones in use
+        [FilePath || FilePath <- FileList,
+            re:run(FilePath, RegExp, [{capture, none}]) =:= nomatch]
+    end,
 
-    % filter out the ones in use
-    DeleteFiles = [FilePath
-           || FilePath <- FileList,
-              re:run(FilePath, RegExp, [{capture, none}]) =:= nomatch],
     % delete unused files
     ?LOG_DEBUG("deleting unused view index files: ~p",[DeleteFiles]),
     RootDir = couch_config:get("couchdb", "view_index_dir"),
@@ -108,7 +114,9 @@ cleanup_index_files(Db) ->
 list_index_files(Db) ->
     % call server to fetch the index files
     RootDir = couch_config:get("couchdb", "view_index_dir"),
-    filelib:wildcard(RootDir ++ "/." ++ ?b2l(couch_db:name(Db)) ++ "_design"++"/*").
+    filelib:wildcard(
+        RootDir ++ "/." ++ ?b2l(couch_db:name(Db)) ++ "_design"++"/*.view"
+    ).
 
 
 get_row_count(#view{btree=Bt}) ->
@@ -247,8 +255,16 @@ fold_fun(Fun, [KV|Rest], {KVReds, Reds}, Acc) ->
 
 fold(#view{btree=Btree}, Fun, Acc, Options) ->
     WrapperFun =
-        fun(KV, Reds, Acc2) ->
-            fold_fun(Fun, expand_dups([KV],[]), Reds, Acc2)
+        fun(visit, KV, Reds, Acc2) ->
+            fold_fun(Fun, expand_dups([KV],[]), Reds, Acc2);
+           (traverse, LK, Red, Acc2)
+              when is_function(Fun, 4) ->
+                Fun(traverse, LK, Red, Acc2);
+           (traverse, _LK, Red, {_, Skip, _, _} = Acc2)
+              when Skip >= element(1, Red) ->
+                {skip, setelement(2, Acc2, Skip - element(1, Red))};
+           (traverse, _, _, Acc2) ->
+                {ok, Acc2}
         end,
     {ok, _LastReduce, _AccResult} = couch_btree:fold(Btree, WrapperFun, Acc, Options).
 
@@ -262,17 +278,29 @@ init([]) ->
             exit(Self, config_change)
         end),
 
+    ets:new(couch_groups_by_db, [bag, protected, named_table]),
+    ets:new(group_servers_by_sig, [set, protected, named_table]),
+    ets:new(couch_groups_by_updater, [set, private, named_table]),
+
     couch_db_update_notifier:start_link(
         fun({deleted, DbName}) ->
             gen_server:cast(couch_view, {reset_indexes, DbName});
         ({created, DbName}) ->
             gen_server:cast(couch_view, {reset_indexes, DbName});
+        ({ddoc_updated, {DbName, DDocId}}) ->
+            lists:foreach(
+                fun({_DbName, {_DDocId, Sig}}) ->
+                    case ets:lookup(group_servers_by_sig, {DbName, Sig}) of
+                    [{_, GroupPid}] ->
+                        (catch gen_server:cast(GroupPid, ddoc_updated));
+                    [] ->
+                        ok
+                    end
+                end,
+                ets:match_object(couch_groups_by_db, {DbName, {DDocId, '$1'}}));
         (_Else) ->
             ok
         end),
-    ets:new(couch_groups_by_db, [bag, private, named_table]),
-    ets:new(group_servers_by_sig, [set, protected, named_table]),
-    ets:new(couch_groups_by_updater, [set, private, named_table]),
     process_flag(trap_exit, true),
     ok = couch_file:init_delete_dir(RootDir),
     {ok, #server{root_dir=RootDir}}.
@@ -312,23 +340,23 @@ new_group(Root, DbName, #group{name=GroupId, sig=Sig} = Group) ->
     case (catch couch_view_group:start_link({Root, DbName, Group})) of
     {ok, NewPid} ->
         unlink(NewPid),
-        exit({DbName, Sig, {ok, NewPid}});
+        exit({DbName, GroupId, Sig, {ok, NewPid}});
     {error, invalid_view_seq} ->
         ok = gen_server:call(couch_view, {reset_indexes, DbName}),
         new_group(Root, DbName, Group);
     Error ->
-        exit({DbName, Sig, Error})
+        exit({DbName, GroupId, Sig, Error})
     end.
 
 do_reset_indexes(DbName, Root) ->
     % shutdown all the updaters and clear the files, the db got changed
     Names = ets:lookup(couch_groups_by_db, DbName),
     lists:foreach(
-        fun({_DbName, Sig}) ->
+        fun({_DbName, {DDocId, Sig}}) ->
             ?LOG_DEBUG("Killing update process for view group ~s. in database ~s.", [Sig, DbName]),
             [{_, Pid}] = ets:lookup(group_servers_by_sig, {DbName, Sig}),
             couch_util:shutdown_sync(Pid),
-            delete_from_ets(Pid, DbName, Sig)
+            delete_from_ets(Pid, DbName, DDocId, Sig)
         end, Names),
     delete_index_dir(Root, DbName),
     RootDelDir = couch_config:get("couchdb", "view_index_dir"),
@@ -343,29 +371,33 @@ handle_info({'EXIT', FromPid, Reason}, Server) ->
             exit(Reason);
         true -> ok
         end;
-    [{_, {DbName, GroupId}}] ->
-        delete_from_ets(FromPid, DbName, GroupId)
+    [{_, {DbName, Sig}}] ->
+        [{DbName, {DDocId, Sig}}] = ets:match_object(
+            couch_groups_by_db, {DbName, {'$1', Sig}}),
+        delete_from_ets(FromPid, DbName, DDocId, Sig)
     end,
     {noreply, Server};
 
-handle_info({'DOWN', _, _, _, {DbName, Sig, Reply}}, Server) ->
+handle_info({'DOWN', _, _, _, {DbName, DDocId, Sig, Reply}}, Server) ->
     [{_, WaitList}] = ets:lookup(group_servers_by_sig, {DbName, Sig}),
     [gen_server:reply(From, Reply) || From <- WaitList],
     case Reply of {ok, NewPid} ->
         link(NewPid),
-        add_to_ets(NewPid, DbName, Sig);
-     _ -> ok end,
+        add_to_ets(NewPid, DbName, DDocId, Sig);
+     _ ->
+        ets:delete(group_servers_by_sig, {DbName, Sig})
+    end,
     {noreply, Server}.
 
-add_to_ets(Pid, DbName, Sig) ->
+add_to_ets(Pid, DbName, DDocId, Sig) ->
     true = ets:insert(couch_groups_by_updater, {Pid, {DbName, Sig}}),
     true = ets:insert(group_servers_by_sig, {{DbName, Sig}, Pid}),
-    true = ets:insert(couch_groups_by_db, {DbName, Sig}).
+    true = ets:insert(couch_groups_by_db, {DbName, {DDocId, Sig}}).
 
-delete_from_ets(Pid, DbName, Sig) ->
+delete_from_ets(Pid, DbName, DDocId, Sig) ->
     true = ets:delete(couch_groups_by_updater, Pid),
     true = ets:delete(group_servers_by_sig, {DbName, Sig}),
-    true = ets:delete_object(couch_groups_by_db, {DbName, Sig}).
+    true = ets:delete_object(couch_groups_by_db, {DbName, {DDocId, Sig}}).
 
 code_change(_OldVsn, State, _Extra) ->
     {ok, State}.
@@ -394,7 +426,7 @@ nuke_dir(RootDelDir, Dir) ->
 
 % keys come back in the language of btree - tuples.
 less_json_ids({JsonA, IdA}, {JsonB, IdB}) ->
-    case less_json0(JsonA, JsonB) of
+    case couch_ejson_compare:less(JsonA, JsonB) of
     0 ->
         IdA < IdB;
     Result ->
@@ -402,59 +434,4 @@ less_json_ids({JsonA, IdA}, {JsonB, IdB}) ->
     end.
 
 less_json(A,B) ->
-    less_json0(A,B) < 0.
-
-less_json0(A,A)                                 -> 0;
-
-less_json0(A,B) when is_atom(A), is_atom(B)     -> atom_sort(A) - atom_sort(B);
-less_json0(A,_) when is_atom(A)                 -> -1;
-less_json0(_,B) when is_atom(B)                 -> 1;
-
-less_json0(A,B) when is_number(A), is_number(B) -> A - B;
-less_json0(A,_) when is_number(A)               -> -1;
-less_json0(_,B) when is_number(B)               -> 1;
-
-less_json0(A,B) when is_binary(A), is_binary(B) -> couch_util:collate(A,B);
-less_json0(A,_) when is_binary(A)               -> -1;
-less_json0(_,B) when is_binary(B)               -> 1;
-
-less_json0(A,B) when is_list(A), is_list(B)     -> less_list(A,B);
-less_json0(A,_) when is_list(A)                 -> -1;
-less_json0(_,B) when is_list(B)                 -> 1;
-
-less_json0({A},{B}) when is_list(A), is_list(B) -> less_props(A,B);
-less_json0({A},_) when is_list(A)               -> -1;
-less_json0(_,{B}) when is_list(B)               -> 1.
-
-atom_sort(null) -> 1;
-atom_sort(false) -> 2;
-atom_sort(true) -> 3.
-
-less_props([], [_|_]) ->
-    -1;
-less_props(_, []) ->
-    1;
-less_props([{AKey, AValue}|RestA], [{BKey, BValue}|RestB]) ->
-    case couch_util:collate(AKey, BKey) of
-    0 ->
-        case less_json0(AValue, BValue) of
-        0 ->
-            less_props(RestA, RestB);
-        Result ->
-            Result
-        end;
-    Result ->
-        Result
-    end.
-
-less_list([], [_|_]) ->
-    -1;
-less_list(_, []) ->
-    1;
-less_list([A|RestA], [B|RestB]) ->
-    case less_json0(A,B) of
-    0 ->
-        less_list(RestA, RestB);
-    Result ->
-        Result
-    end.
+    couch_ejson_compare:less(A, B) < 0.
