@@ -15,7 +15,11 @@
 
 %% API
 -export([start_link/1, request_group/2, request_group_info/1]).
--export([open_db_group/2, open_temp_group/5, design_doc_to_view_group/1,design_root/2]).
+-export([open_db_group/2, open_temp_group/5, design_doc_to_view_group/1]).
+-export([design_root/2, index_file_name/3]).
+
+%% Exports for the compactor
+-export([get_index_header_data/1]).
 
 %% gen_server callbacks
 -export([init/1, handle_call/3, handle_cast/2, handle_info/2,
@@ -32,7 +36,8 @@
     compactor_pid=nil,
     waiting_commit=false,
     waiting_list=[],
-    ref_counter=nil
+    ref_counter=nil,
+    shutdown=false
 }).
 
 % api methods
@@ -78,7 +83,7 @@ start_link(InitArgs) ->
 init({{_, DbName, _} = InitArgs, ReturnPid, Ref}) ->
     process_flag(trap_exit, true),
     try prepare_group(InitArgs, false) of
-    {ok, #group{db=Db, fd=Fd, current_seq=Seq}=Group} ->
+    {ok, Db, #group{fd=Fd, current_seq=Seq}=Group} ->
         case Seq > couch_db:get_update_seq(Db) of
         true ->
             ReturnPid ! {Ref, self(), {error, invalid_view_seq}},
@@ -90,7 +95,7 @@ init({{_, DbName, _} = InitArgs, ReturnPid, Ref}) ->
             {ok, #group_state{
                     db_name=DbName,
                     init_args=InitArgs,
-                    group=Group#group{db=nil},
+                    group=Group,
                     ref_counter=RefCounter}}
         end;
     Error ->
@@ -124,14 +129,11 @@ handle_call({request_group, RequestSeq}, From,
             updater_pid=nil,
             waiting_list=WaitList
             }=State) when RequestSeq > Seq ->
-    {ok, Db} = couch_db:open_int(DbName, []),
-    Group2 = Group#group{db=Db},
     Owner = self(),
-    Pid = spawn_link(fun()-> couch_view_updater:update(Owner, Group2) end),
+    Pid = spawn_link(fun()-> couch_view_updater:update(Owner, Group, DbName) end),
 
     {noreply, State#group_state{
         updater_pid=Pid,
-        group=Group2,
         waiting_list=[{From,RequestSeq}|WaitList]
         }, infinity};
 
@@ -154,9 +156,9 @@ handle_call({request_group, RequestSeq}, From,
 
 handle_call(request_group_info, _From, State) ->
     GroupInfo = get_group_info(State),
-    {reply, {ok, GroupInfo}, State}.
+    {reply, {ok, GroupInfo}, State};
 
-handle_cast({start_compact, CompactFun}, #group_state{compactor_pid=nil}
+handle_call({start_compact, CompactFun}, _From, #group_state{compactor_pid=nil}
         = State) ->
     #group_state{
         group = #group{name = GroupId, sig = GroupSig} = Group,
@@ -166,17 +168,23 @@ handle_cast({start_compact, CompactFun}, #group_state{compactor_pid=nil}
     {ok, Db} = couch_db:open_int(DbName, []),
     {ok, Fd} = open_index_file(compact, RootDir, DbName, GroupSig),
     NewGroup = reset_file(Db, Fd, DbName, Group),
-    Pid = spawn_link(fun() -> CompactFun(Group, NewGroup) end),
-    {noreply, State#group_state{compactor_pid = Pid}};
-handle_cast({start_compact, _}, State) ->
+    couch_db:close(Db),
+    unlink(Fd),
+    Pid = spawn_link(fun() ->
+        link(Fd),
+        CompactFun(Group, NewGroup, DbName),
+        unlink(Fd)
+    end),
+    {reply, {ok, Pid}, State#group_state{compactor_pid = Pid}};
+handle_call({start_compact, _}, _From, State) ->
     %% compact already running, this is a no-op
-    {noreply, State};
+    {reply, {ok, State#group_state.compactor_pid}, State};
 
-handle_cast({compact_done, #group{current_seq=NewSeq} = NewGroup},
+handle_call({compact_done, #group{current_seq=NewSeq} = NewGroup}, _From,
         #group_state{group = #group{current_seq=OldSeq}} = State)
         when NewSeq >= OldSeq ->
     #group_state{
-        group = #group{name=GroupId, fd=OldFd, sig=GroupSig} = Group,
+        group = #group{name=GroupId, fd=OldFd, sig=GroupSig},
         init_args = {RootDir, DbName, _},
         updater_pid = UpdaterPid,
         compactor_pid = CompactorPid,
@@ -195,7 +203,7 @@ handle_cast({compact_done, #group{current_seq=NewSeq} = NewGroup},
         unlink(UpdaterPid),
         exit(UpdaterPid, view_compaction_complete),
         Owner = self(),
-        spawn_link(fun()-> couch_view_updater:update(Owner, NewGroup) end);
+        spawn_link(fun()-> couch_view_updater:update(Owner, NewGroup, DbName) end);
     true ->
         nil
     end,
@@ -206,63 +214,88 @@ handle_cast({compact_done, #group{current_seq=NewSeq} = NewGroup},
     unlink(OldFd),
     couch_ref_counter:drop(RefCounter),
     {ok, NewRefCounter} = couch_ref_counter:start([NewGroup#group.fd]),
-    case Group#group.db of
-        nil -> ok;
-        Else -> couch_db:close(Else)
-    end,
-
-    case NewGroup#group.db of
-        nil -> ok;
-        _ -> couch_db:close(NewGroup#group.db)
-    end,
 
     self() ! delayed_commit,
-    {noreply, State#group_state{
-        group=NewGroup#group{db = nil},
+    {reply, ok, State#group_state{
+        group=NewGroup,
         ref_counter=NewRefCounter,
         compactor_pid=nil,
         updater_pid=NewUpdaterPid
     }};
-handle_cast({compact_done, NewGroup}, State) ->
+handle_call({compact_done, NewGroup}, _From, State) ->
     #group_state{
         group = #group{name = GroupId, current_seq = CurrentSeq},
         init_args={_RootDir, DbName, _}
     } = State,
     ?LOG_INFO("View index compaction still behind for ~s ~s -- current: ~p " ++
         "compact: ~p", [DbName, GroupId, CurrentSeq, NewGroup#group.current_seq]),
-    couch_db:close(NewGroup#group.db),
-    Pid = spawn_link(fun() ->
-        {ok, Db} = couch_db:open_int(DbName, []),
-        {_,Ref} = erlang:spawn_monitor(fun() ->
-            couch_view_updater:update(nil, NewGroup#group{db = Db})
-        end),
-        receive
-            {'DOWN', Ref, _, _, {new_group, NewGroup2}} ->
-                couch_db:close(Db),
-                #group{name=GroupId} = NewGroup2,
-                Pid2 = couch_view:get_group_server(DbName, GroupId),
-                gen_server:cast(Pid2, {compact_done, NewGroup2#group{db = nil}})
-        end
-    end),
-    {noreply, State#group_state{compactor_pid = Pid}};
+    {reply, update, State};
+
+handle_call(cancel_compact, _From, #group_state{compactor_pid = nil} = State) ->
+    {reply, ok, State};
+handle_call(cancel_compact, _From, #group_state{compactor_pid = Pid} = State) ->
+    unlink(Pid),
+    exit(Pid, kill),
+    #group_state{
+        group = #group{sig=GroupSig},
+        init_args = {RootDir, DbName, _}
+    } = State,
+    CompactFile = index_file_name(compact, RootDir, DbName, GroupSig),
+    ok = couch_file:delete(RootDir, CompactFile),
+    {reply, ok, State#group_state{compactor_pid = nil}}.
+
 
 handle_cast({partial_update, Pid, NewGroup}, #group_state{updater_pid=Pid}
         = State) ->
     #group_state{
         db_name = DbName,
-        waiting_commit = WaitingCommit
+        waiting_commit = WaitingCommit,
+        group = Group
     } = State,
     NewSeq = NewGroup#group.current_seq,
-    ?LOG_INFO("checkpointing view update at seq ~p for ~s ~s", [NewSeq,
-        DbName, NewGroup#group.name]),
-    if not WaitingCommit ->
-        erlang:send_after(1000, self(), delayed_commit);
-    true -> ok
-    end,
-    {noreply, State#group_state{group=NewGroup, waiting_commit=true}};
+    CheckpointInterval = list_to_integer(
+        couch_config:get("couchdb","view_checkpoint_interval","100")),
+    case NewSeq > CheckpointInterval + Group#group.current_seq of
+    true ->
+        ?LOG_INFO("checkpointing view update at seq ~p for ~s ~s", [NewSeq,
+            DbName, NewGroup#group.name]),
+        if not WaitingCommit ->
+            erlang:send_after(1000, self(), delayed_commit);
+        true -> ok
+        end,
+        {noreply, State#group_state{group=NewGroup, waiting_commit=true}};
+    false ->
+        {noreply, State}
+    end;
 handle_cast({partial_update, _, _}, State) ->
     %% message from an old (probably pre-compaction) updater; ignore
-    {noreply, State}.
+    {noreply, State};
+handle_cast(ddoc_updated, State) ->
+    #group_state{
+        db_name = DbName,
+        waiting_list = Waiters,
+        group = #group{name = DDocId, sig = CurSig}
+    } = State,
+    {ok, Db} = couch_db:open_int(DbName, []),
+    case couch_db:open_doc(Db, DDocId, [ejson_body]) of
+    {not_found, deleted} ->
+        NewSig = nil;
+    {ok, DDoc} ->
+        #group{sig = NewSig} = design_doc_to_view_group(DDoc)
+    end,
+    couch_db:close(Db),
+    case NewSig of
+    CurSig ->
+        {noreply, State#group_state{shutdown = false}};
+    _ ->
+        case Waiters of
+        [] ->
+            {stop, normal, State};
+        _ ->
+            {noreply, State#group_state{shutdown = true}}
+        end
+    end.
+
 
 handle_info(delayed_commit, #group_state{db_name=DbName,group=Group}=State) ->
     {ok, Db} = couch_db:open_int(DbName, []),
@@ -283,44 +316,46 @@ handle_info(delayed_commit, #group_state{db_name=DbName,group=Group}=State) ->
         {noreply, State#group_state{waiting_commit=true}}
     end;
 
-handle_info({'EXIT', FromPid, {new_group, #group{db=Db}=Group}},
+handle_info({'EXIT', FromPid, {new_group, Group}},
         #group_state{db_name=DbName,
             updater_pid=UpPid,
             ref_counter=RefCounter,
             waiting_list=WaitList,
+            shutdown=Shutdown,
             waiting_commit=WaitingCommit}=State) when UpPid == FromPid ->
-    ok = couch_db:close(Db),
     if not WaitingCommit ->
         erlang:send_after(1000, self(), delayed_commit);
     true -> ok
     end,
     case reply_with_group(Group, WaitList, [], RefCounter) of
     [] ->
-        {noreply, State#group_state{waiting_commit=true, waiting_list=[],
-                group=Group#group{db=nil}, updater_pid=nil}};
+        case Shutdown of
+        true ->
+            {stop, normal, State};
+        false ->
+            {noreply, State#group_state{waiting_commit=true, waiting_list=[],
+                group=Group, updater_pid=nil}}
+        end;
     StillWaiting ->
         % we still have some waiters, reopen the database and reupdate the index
-        {ok, Db2} = couch_db:open_int(DbName, []),
-        Group2 = Group#group{db=Db2},
         Owner = self(),
-        Pid = spawn_link(fun() -> couch_view_updater:update(Owner, Group2) end),
+        Pid = spawn_link(fun() -> couch_view_updater:update(Owner, Group, DbName) end),
         {noreply, State#group_state{waiting_commit=true,
-                waiting_list=StillWaiting, group=Group2, updater_pid=Pid}}
+                waiting_list=StillWaiting, updater_pid=Pid}}
     end;
 handle_info({'EXIT', _, {new_group, _}}, State) ->
     %% message from an old (probably pre-compaction) updater; ignore
     {noreply, State};
 
-handle_info({'EXIT', FromPid, reset},
-        #group_state{
-            init_args=InitArgs,
-            updater_pid=UpPid,
-            group=Group}=State) when UpPid == FromPid ->
-    ok = couch_db:close(Group#group.db),
+handle_info({'EXIT', UpPid, reset},
+        #group_state{init_args=InitArgs, updater_pid=UpPid} = State) ->
     case prepare_group(InitArgs, true) of
-    {ok, ResetGroup} ->
+    {ok, Db, ResetGroup} ->
         Owner = self(),
-        Pid = spawn_link(fun()-> couch_view_updater:update(Owner, ResetGroup) end),
+        couch_db:close(Db),
+        Pid = spawn_link(fun() ->
+            couch_view_updater:update(Owner, ResetGroup, Db#db.name)
+        end),
         {noreply, State#group_state{
                 updater_pid=Pid,
                 group=ResetGroup}};
@@ -330,7 +365,7 @@ handle_info({'EXIT', FromPid, reset},
 handle_info({'EXIT', _, reset}, State) ->
     %% message from an old (probably pre-compaction) updater; ignore
     {noreply, State};
-    
+
 handle_info({'EXIT', _FromPid, normal}, State) ->
     {noreply, State};
 
@@ -386,21 +421,22 @@ prepare_group({RootDir, DbName, #group{sig=Sig}=Group}, ForceReset)->
         {ok, Fd} ->
             if ForceReset ->
                 % this can happen if we missed a purge
-                {ok, reset_file(Db, Fd, DbName, Group)};
+                {ok, Db, reset_file(Db, Fd, DbName, Group)};
             true ->
-                % 09 UPGRADE CODE
-                ok = couch_file:upgrade_old_header(Fd, <<$r, $c, $k, 0>>),
                 case (catch couch_file:read_header(Fd)) of
                 {ok, {Sig, HeaderInfo}} ->
                     % sigs match!
-                    {ok, init_group(Db, Fd, Group, HeaderInfo)};
+                    {ok, Db, init_group(Db, Fd, Group, HeaderInfo)};
                 _ ->
                     % this happens on a new file
-                    {ok, reset_file(Db, Fd, DbName, Group)}
+                    {ok, Db, reset_file(Db, Fd, DbName, Group)}
                 end
             end;
-        Error ->
-            catch delete_index_file(RootDir, DbName, Sig),
+        {error, Reason} = Error ->
+            ?LOG_ERROR("Failed to open view file '~s': ~s", [
+                index_file_name(RootDir, DbName, Sig),
+                file:format_error(Reason)
+            ]),
             Error
         end;
     Else ->
@@ -425,6 +461,10 @@ hex_sig(GroupSig) ->
 design_root(RootDir, DbName) ->
     RootDir ++ "/." ++ ?b2l(DbName) ++ "_design/".
 
+index_file_name(RootDir, DBName, Pid) when is_pid(Pid) ->
+    {ok, GroupInfo} = request_group_info(Pid),
+    GroupSig = couch_util:from_hex(couch_util:get_value(signature, GroupInfo)),
+    index_file_name(RootDir, DBName, GroupSig);
 index_file_name(RootDir, DbName, GroupSig) ->
     design_root(RootDir, DbName) ++ hex_sig(GroupSig) ++".view".
 
@@ -502,9 +542,9 @@ sort_lib([{LName, LCode}|Rest], LAcc) ->
     sort_lib(Rest, [{LName, LCode}|LAcc]).
 
 open_db_group(DbName, GroupId) ->
-    case couch_db:open_int(DbName, []) of
+    case couch_db:open_int(DbName, [{user_ctx, #user_ctx{roles=[<<"_admin">>]}}]) of
     {ok, Db} ->
-        case couch_db:open_doc(Db, GroupId) of
+        case couch_db:open_doc(Db, GroupId, [ejson_body]) of
         {ok, Doc} ->
             couch_db:close(Db),
             {ok, design_doc_to_view_group(Doc)};
@@ -527,15 +567,18 @@ get_group_info(State) ->
     #group{
         fd = Fd,
         sig = GroupSig,
+        id_btree = Btree,
         def_lang = Lang,
         current_seq=CurrentSeq,
-        purge_seq=PurgeSeq
+        purge_seq=PurgeSeq,
+        views = Views
     } = Group,
     {ok, Size} = couch_file:bytes(Fd),
     [
         {signature, ?l2b(hex_sig(GroupSig))},
         {language, Lang},
         {disk_size, Size},
+        {data_size, view_group_data_size(Btree, Views)},
         {updater_running, UpdaterPid /= nil},
         {compact_running, CompactorPid /= nil},
         {waiting_commit, WaitingCommit},
@@ -544,6 +587,21 @@ get_group_info(State) ->
         {purge_seq, PurgeSeq}
     ].
 
+view_group_data_size(MainBtree, Views) ->
+    lists:foldl(
+        fun(#view{btree = Btree}, Acc) ->
+            sum_btree_sizes(Acc, couch_btree:size(Btree))
+        end,
+        couch_btree:size(MainBtree),
+        Views).
+
+sum_btree_sizes(nil, _) ->
+    null;
+sum_btree_sizes(_, nil) ->
+    null;
+sum_btree_sizes(Size1, Size2) ->
+    Size1 + Size2.
+
 % maybe move to another module
 design_doc_to_view_group(#doc{id=Id,body={Fields}}) ->
     Language = couch_util:get_value(<<"language">>, Fields, <<"javascript">>),
@@ -582,7 +640,7 @@ design_doc_to_view_group(#doc{id=Id,body={Fields}}) ->
 
 reset_group(#group{views=Views}=Group) ->
     Views2 = [View#view{btree=nil} || View <- Views],
-    Group#group{db=nil,fd=nil,query_server=nil,current_seq=0,
+    Group#group{fd=nil,query_server=nil,current_seq=0,
             id_btree=nil,views=Views2}.
 
 reset_file(Db, Fd, DbName, #group{sig=Sig,name=Name} = Group) ->
@@ -591,9 +649,6 @@ reset_file(Db, Fd, DbName, #group{sig=Sig,name=Name} = Group) ->
     ok = couch_file:write_header(Fd, {Sig, nil}),
     init_group(Db, Fd, reset_group(Group), nil).
 
-delete_index_file(RootDir, DbName, GroupSig) ->
-    couch_file:delete(RootDir, index_file_name(RootDir, DbName, GroupSig)).
-
 init_group(Db, Fd, #group{views=Views}=Group, nil) ->
     init_group(Db, Fd, Group,
         #index_header{seq=0, purge_seq=couch_db:get_purge_seq(Db),
@@ -607,7 +662,8 @@ init_group(Db, Fd, #group{def_lang=Lang,views=Views}=
         (State) -> {State, 0, 0}
     end,
     ViewStates2 = lists:map(StateUpdate, ViewStates),
-    {ok, IdBtree} = couch_btree:open(IdBtreeState, Fd),
+    {ok, IdBtree} = couch_btree:open(
+        IdBtreeState, Fd, [{compression, Db#db.compression}]),
     Views2 = lists:zipwith(
         fun({BTState, USeq, PSeq}, #view{reduce_funs=RedFuns,options=Options}=View) ->
             FunSrcs = [FunSrc || {_Name, FunSrc} <- RedFuns],
@@ -625,7 +681,7 @@ init_group(Db, Fd, #group{def_lang=Lang,views=Views}=
                         UserReds),
                     {Count, Reduced}
                 end,
-            
+
             case couch_util:get_value(<<"collation">>, Options, <<"default">>) of
             <<"default">> ->
                 Less = fun couch_view:less_json_ids/2;
@@ -633,10 +689,11 @@ init_group(Db, Fd, #group{def_lang=Lang,views=Views}=
                 Less = fun(A,B) -> A < B end
             end,
             {ok, Btree} = couch_btree:open(BTState, Fd,
-                    [{less, Less}, {reduce, ReduceFun}]
+                    [{less, Less}, {reduce, ReduceFun},
+                        {compression, Db#db.compression}]
             ),
             View#view{btree=Btree, update_seq=USeq, purge_seq=PSeq}
         end,
         ViewStates2, Views),
-    Group#group{db=Db, fd=Fd, current_seq=Seq, purge_seq=PurgeSeq,
+    Group#group{fd=Fd, current_seq=Seq, purge_seq=PurgeSeq,
         id_btree=IdBtree, views=Views2}.
