@@ -40,8 +40,8 @@
 %% @type response(). A mochiweb_response parameterized module instance.
 %% @type ioheaders() = headers() | [{key(), value()}].
 
-% 10 second default idle timeout
--define(IDLE_TIMEOUT, 10000).
+% 5 minute default idle timeout
+-define(IDLE_TIMEOUT, 300000).
 
 % Maximum recv_body() length of 1MB
 -define(MAX_RECV_BODY, (1024*1024)).
@@ -382,8 +382,8 @@ ok({ContentType, ResponseHeaders, Body}) ->
 %% @doc Return true if the connection must be closed. If false, using
 %%      Keep-Alive should be safe.
 should_close() ->
-    ForceClose = erlang:get(mochiweb_request_force_close) =/= undefined,
-    DidNotRecv = erlang:get(mochiweb_request_recv) =:= undefined,
+    ForceClose = erlang:get(?SAVE_FORCE_CLOSE) =/= undefined,
+    DidNotRecv = erlang:get(?SAVE_RECV) =:= undefined,
     ForceClose orelse Version < {1, 0}
         %% Connection: close
         orelse get_header_value("connection") =:= "close"
@@ -405,6 +405,7 @@ cleanup() ->
                        ?SAVE_PATH,
                        ?SAVE_RECV,
                        ?SAVE_BODY,
+                       ?SAVE_BODY_LENGTH,
                        ?SAVE_POST,
                        ?SAVE_COOKIE,
                        ?SAVE_FORCE_CLOSE]],
